from pathlib import Path

import pyarrow as pa  # type: ignore
import pyarrow.compute as pc  # type: ignore
import pyarrow.json as pj  # type: ignore
from typing_extensions import Self
from upath import UPath

from .resource import CoreResource


class CoreManifest(CoreResource):
    """In-memory representation of a serialized package manifest."""

    # TODO: cache Blobs to avoid repeated lookups
    # TODO: improve/replace child(key) handling

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        with path.open(mode="rb") as fi:
            self.table = pj.read_json(fi)
        self.body = self.setup()
        self.name_col = self.cf.get_str("schema/name", "logical_key")
        self.loc_col = self.cf.get_str("schema/location", "physical_keys")

    def setup(self) -> pa.Table:
        first = self.table.take([0]).to_pydict()
        headers = self.cf.get_dict("schema/headers")
        keys = list(headers.keys())
        for key in keys:
            setattr(self, key, first[key][0])
        return self.table.drop_columns(keys).slice(1)

    def child_row(self, key: str) -> dict:
        """Return the dict for a child resource."""
        # TODO: cache to avoid continually re-calcluating
        rows = self.body.filter(pc.field(self.name_col) == key)
        if rows.num_rows == 0:
            raise KeyError(f"Key [{key}] not found in {self.name_col} of {self.path}")
        return rows.to_pydict()

    def child_path(self, key: str) -> Path:
        """Return the path for a child resource."""
        row = self.child_row(key)
        loc = row[self.loc_col][0][0]
        return UPath(loc)

    def child_args(self, key: str) -> dict:
        """Return the parameters for a child resource."""
        row = self.child_row(key)
        return {"row": row, "parent": self}

    def list(self) -> list[Self]:
        """List all child resources."""
        names = self.body.column(self.name_col).to_pylist()
        return [self.child(self.child_path(x), x) for x in names]



from __future__ import annotations

from pathlib import Path

import quiltcore
from typing_extensions import Self

from .config import CoreConfig


class CoreResource:
    @staticmethod
    def ClassFromName(name: str) -> type:
        """Return a class from a string."""
        return getattr(quiltcore, name)

    """Generic resource class."""

    def __init__(self, path: Path, **kwargs):
        self.cf = CoreConfig()
        self.name = self.__class__.__name__
        rkey = f"resources/{self.name}"
        self.params = self.cf.get_dict(rkey)
        self.path = path
        _child = self.param("child", "CoreResource")
        self.klass = CoreResource.ClassFromName(_child)
        self.glob = self.param("glob", "*")

    def __repr__(self):
        return f"<{self.__class__} {self.path}>"

    def __str__(self):
        return str(self.path)

    def param(self, key: str, default: str) -> str:
        """Return a param."""
        return self.params[key] if key in self.params else default  # type: ignore

    def child_args(self, key: str) -> dict:
        """Return the parameters for a child resource."""
        return {}

    def child(self, path: Path, key: str = ""):
        """Return a child resource."""
        return self.klass(path, **self.child_args(key))

    def child_path(self, key: str) -> Path:
        """Return the path for a child resource."""
        return self.path / key

    def list_gen(self):
        return self.path.glob(self.glob)

    def list(self) -> list[Self]:
        """List all child resources."""
        return [self.child(x) for x in self.list_gen()]

    def get(self, key: str) -> Self:
        """Get a child resource by name."""
        path = self.child_path(key)
        if not path.exists():
            raise KeyError(f"Key {key} not found in {self.path}")
        return self.child(path, key)

    def put(self, dest: Path) -> Path:
        """Put a resource into dest. Return the new path"""
        raise NotImplementedError



from pathlib import Path

from .resource import CoreResource


class CoreValues(CoreResource):

    """Storage for dereferenced Names"""

    def __init__(self, values: Path, parent: CoreResource):
        super().__init__(values)



from pathlib import Path

from .resource import CoreResource
from .values import CoreValues


class CoreRegistry(CoreResource):
    """
    Registry of Names and parent of Values
    `list` and `get` return CoreName with parent of CoreValues
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        base = path / self.cf.get_path("dirs/config")
        self.path = base / self.cf.get_path("dirs/names")
        values = base / self.cf.get_path("dirs/values")
        self.values = CoreValues(values, self)

    def child_args(self, key: str) -> dict:
        return {"values": self.values}



from pathlib import Path

from multiformats import multihash

from .manifest import CoreManifest
from .resource import CoreResource


class CoreBlob(CoreResource):
    """Storage for dereferenced Names"""

    MH_PREFIX = "1220"

    @staticmethod
    def FromKeyInManifest(key: str, manifest: CoreManifest) -> "CoreBlob":
        path = manifest.child_path(key)
        row = manifest.child_row(key)
        return CoreBlob(path, parent=manifest, row=row)

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        self.parent = kwargs["parent"]
        self.row = kwargs["row"]
        if len(self.row) > 0:
            self.setup(self.row)

    def setup(self, row: dict):
        columns = self.cf.get_dict("schema/columns")
        for key, type in columns.items():
            value = row[key][0]
            if isinstance(value, list):
                value = value[0]
                key = key.rstrip("s")
            if type == "int64":
                value = int(value)
            setattr(self, key, value)
            if key == "hash":
                self.setup_hash(value)  # type: ignore

    def setup_hash(self, opt: dict):
        self.hash_value = opt["value"]
        hash_key = f'multihash/{opt["type"]}'
        self.hash_type = self.cf.get_str(hash_key)
        self.hash_digest = multihash.get(self.hash_type)

    def digest(self, bstring: bytes) -> str:
        digest = self.hash_digest.digest(bstring)
        hex = digest.hex()
        return hex.strip(CoreBlob.MH_PREFIX)

    def name(self):
        return self.row[self.parent.name_col]  # type: ignore

    def location(self):
        return self.row[self.parent.loc_col]  # type: ignore

    def put(self, dest: Path) -> Path:
        """Put a resource into dest. Return the new path"""
        dest.write_bytes(self.path.read_bytes())  # for binary files
        return dest

    def verify(self, bstring: bytes) -> bool:
        """Verify that bytes match the hash"""
        digest = self.digest(bstring)
        print(digest)
        print(self.hash_value)
        return digest == self.hash_value



from pathlib import Path

from un_yaml import UnYaml  # type: ignore


class CoreConfig(UnYaml):
    CONFIG_FILE = "config.yaml"

    @classmethod
    def DefaultConfig(cls) -> dict:
        return UnYaml.LoadYaml(cls.CONFIG_FILE, __package__)

    def __init__(self, yaml_data: dict = {}) -> None:
        data = yaml_data if len(yaml_data) > 0 else CoreConfig.DefaultConfig()
        super().__init__(data)

    def get_str(self, key: str, default="") -> str:
        return super().get(key) or default

    def get_path(self, key: str) -> Path:
        str_path = self.get_str(key, ".")
        return Path(str_path)

    def get_dict(self, key: str) -> dict:
        return self.get(key) or {}



from pathlib import Path

from .resource import CoreResource


class CoreName(CoreResource):
    """Namespace of Manifests by Hash"""

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        self.values = kwargs["values"]

    def hash(self, tag: str = "latest") -> str:
        hash_file = self.path / tag
        return hash_file.read_text()

    def child_path(self, key: str) -> Path:
        """Return the path for a child resource."""
        hash = self.hash(key)
        return self.values.path / hash



from .blob import CoreBlob  # noqa: F401
from .config import CoreConfig  # noqa: F401
from .manifest import CoreManifest  # noqa: F401
from .name import CoreName  # noqa: F401
from .registry import CoreRegistry  # noqa: F401
from .resource import CoreResource  # noqa: F401
from .values import CoreValues  # noqa: F401
